/*
 * Copyright (C) 2004, 2005, 2006 Moritz Orbach <zufall@apfelboymchen.homeunix.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Parts of this file are ripped from Esetroot.c written by Nat Friedman 
 * <ndf@mit.edu> with modifications by Gerald Britton <gbritton@mit.edu> and 
 * Michael Jennings <mej@eterm.org>
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/Xos.h>
#include <Imlib2.h>

#include "config.h"

#include "cfg.h"
#include "command.h"

#ifdef FENCE
#include "efence.h"
#endif

Pixmap CurrentPixmap = None;


void showpixmaps(Display *Xdisplay, Window Xroot)
{
	Atom prop_root = None, prop_esetroot = None, type;
	int format;
	unsigned long length, after;
	unsigned char *data;
	Pixmap p;

	prop_root = XInternAtom(Xdisplay, "_XROOTPMAP_ID", True);
	prop_esetroot = XInternAtom(Xdisplay, "ESETROOT_PMAP_ID", True);

	if (prop_root != None) {
		if (XGetWindowProperty(Xdisplay, Xroot, prop_root, 0L, 1L, False, AnyPropertyType, &type, &format, &length, &after, &data) == Success)
			printf("success!");
		if (data) {
			p = *((Pixmap *)data);
			fprintf(stderr, "xroot   : %p\n", (void *)p);
		} else printf("kein xroot-data\n");
	} else printf("kein xroot\n");

	if (prop_esetroot != None) {
		if (XGetWindowProperty(Xdisplay, Xroot, prop_esetroot, 0L, 1L, False, AnyPropertyType, &type, &format, &length, &after, &data) == Success)
			printf("esuccess!");
		if (data) {
			p = *((Pixmap *)data);
			fprintf(stderr, "esetroot: %p\n", (void *)p);
		} else printf("kein esetroot-data\n");
	} else printf("kein esetroot\n");
	printf("--------\n");
}

/*
 * http://www.eterm.org/docs/view.php?doc=ref#trans
 * http://foo-projects.org/pipermail/xfce4-dev/2003-January/001699.html
 * 
 * In summary:
 * 
 * _XROOTPMAP_ID is the pixmap to be displayed. ESETROOT_PMAP_ID is the
 * pixmap to be destroyed, if any. A program that quits after setting the
 * backdrop will set ESETROOT_PMAP_ID, otherwise it won't, so both types of
 * program are supported.
 *
 * ???:
 * http://mail.gnome.org/archives/wm-spec-list/2002-January/msg00006.html
 * There is no point in checking to see if ESETROOT_PMAP_ID is
 * equal to _XROOTPMAP_ID ... if ESETROOT_PMAP_ID is set, it
 * will point to a pixmap that must be killed before changing
 * ESETROOT_PMAP_ID.
 * - because i have no idea i will do it like stolen from Esetroot
 */
void kill_esetroot(Display *Xdisplay, Window Xroot)
{
	Atom prop_root = None;
	Atom prop_esetroot = None;
	Atom type = None;
	int format;
	unsigned long length, after;
	unsigned char *data_root, *data_esetroot;

	/* First check to see if the properties already exist, and if they are equal */
	prop_root = XInternAtom(Xdisplay, "_XROOTPMAP_ID", True);
	/* WHY is prop_esetroot NEVER != None ??! */
	prop_esetroot = XInternAtom(Xdisplay, "ESETROOT_PMAP_ID", True);

	if (prop_root != None && prop_esetroot != None) {
		XGetWindowProperty(Xdisplay, Xroot, prop_root, 0L, 1L, False, AnyPropertyType, &type, &format, &length, &after, &data_root);
		if (type == XA_PIXMAP) {
			XGetWindowProperty(Xdisplay, Xroot, prop_esetroot, 0L, 1L, False, AnyPropertyType, &type, &format, &length, &after, &data_esetroot);
			if (data_root && data_esetroot && type == XA_PIXMAP && *((Pixmap *) data_root) == *((Pixmap *)data_esetroot)) {
				/*
				 * kill the previous image
				 * If the client has already terminated in either RetainPermanent or RetainTemporary 
				 * mode, all of the client's resources are destroyed.
				 */
				XKillClient(Xdisplay, *((Pixmap *)data_esetroot));
			}
		}
	}
}

void remove_property(Display *Xdisplay, Window Xroot)
{
	/*
	 * important! XInternAtom doesn't seem to touch id if there is no atom 
	 * that took a while.. :(
	 */
	Atom id = None;
	    
	id = XInternAtom (Xdisplay, "ESETROOT_PMAP_ID", True);
	if (id != None) {
		/* kill pixmap first! */
		kill_esetroot(Xdisplay, Xroot);
		XDeleteProperty (Xdisplay, Xroot, id);
	}
}



/*
 * sets ESETROOT_PMAP_ID to _XROOTPMAP_ID
 */
void set_permanent(Display *Xdisplay, Window Xroot)
{
	Atom prop_root, prop_esetroot, type;
	int format;
	unsigned long length, after;
	unsigned char *data_root;
	Pixmap p;

	if (CurrentPixmap == None)
		/* no image set yet */
		return;
	
	/* First check to see if the properties already exist */
	prop_root = XInternAtom(Xdisplay, "_XROOTPMAP_ID", True);
	prop_esetroot = XInternAtom(Xdisplay, "ESETROOT_PMAP_ID", True);

	if (prop_root == None) {
		/* huh? */
		fprintf(stderr, "%s: Could not get my own atom\n", PACKAGE_NAME);
		return;
	}

	if (prop_esetroot == None) {
		/* create */
		prop_esetroot = XInternAtom(Xdisplay, "ESETROOT_PMAP_ID", False);
	}
	else {
		/* some other program set the background */
		kill_esetroot(Xdisplay, Xroot);
	}

	/* get _XROOTPMAP_ID, set ESETROOT_PMAP_ID to _XROOTPMAP_ID */
	XGetWindowProperty(Xdisplay, Xroot, prop_root, 0L, 1L, False, AnyPropertyType, &type, &format, &length, &after, &data_root);
	p = *((Pixmap *)data_root);
	XChangeProperty(Xdisplay, Xroot, prop_esetroot, XA_PIXMAP, 32, PropModeReplace, (unsigned char *) &p, 1);
	/*
	 * leave pixmap in X-server after exit.
	 * Has to be killed by the next program that sets the background!
	 */
	XSetCloseDownMode(Xdisplay, RetainPermanent);
}


/*
 * http://mail.gnome.org/archives/wm-spec-list/2002-January/msg00006.html
 * > Now one problem with the E method is that it may need
 * > a lot of memory (this depends on the memory you have and
 * It takes _no_ extra memory. The pixmap is already referenced
 * by the server, all the E method does is keep around a 
 * XID pointing to the pixmap.
 *
 * Regards,
 */
void set_pixmap_permanent(Display *Xdisplay, Window Xroot, Pixmap p)
{
	/* Xdisplay is the Display, Xroot is the desktop Window, p is the pixmap */
	Atom prop_root = None;
	Atom prop_esetroot = None;

	/* This will locate the property, creating it if it doesn't exist */
	prop_root = XInternAtom(Xdisplay, "_XROOTPMAP_ID", False);
	prop_esetroot = XInternAtom(Xdisplay, "ESETROOT_PMAP_ID", False);

	if (prop_esetroot != None)
		kill_esetroot(Xdisplay, Xroot);

	/* The call above should have created it.  If that failed, we can't continue. */
	if (prop_root == None || prop_esetroot == None) {
		fprintf(stderr, "%s:  Creation of pixmap property failed.\n", PACKAGE_NAME);
		exit(EXIT_FAILURE);
	}
	/* set prop_root and prop_esetroot to the new pixmap */
	XChangeProperty(Xdisplay, Xroot, prop_root, XA_PIXMAP, 32, PropModeReplace, (unsigned char *) &p, 1);
	XChangeProperty(Xdisplay, Xroot, prop_esetroot, XA_PIXMAP, 32, PropModeReplace, (unsigned char *) &p, 1);
	XSetCloseDownMode(Xdisplay, RetainPermanent);
}

void set_pixmap_property(Display *Xdisplay, Window Xroot, Pixmap p)
{
	Atom prop_root;

	prop_root = XInternAtom(Xdisplay, "_XROOTPMAP_ID", True);

	if (prop_root == None)
	{
		/* This will locate the property, creating it if it doesn't exist */
		prop_root = XInternAtom(Xdisplay, "_XROOTPMAP_ID", False);
	}

	/* The call above should have created it.  If that failed, we can't continue. */
	if (prop_root == None) {
		fprintf(stderr, "%s: creation of pixmap property failed.\n", PACKAGE_NAME);
		exit(1);
	}
	/* set prop_root to the new pixmap */
	XChangeProperty(Xdisplay, Xroot, prop_root, XA_PIXMAP, 32, PropModeReplace, (unsigned char *) &p, 1);

}


/* convert dezimal color-value to hex and append to array */
void append_color(int color_dez, char color[7])
{
	/* 2 hex + \0 */
	char temp[3];

	snprintf(temp, 3, "%.2x\n", color_dez);
	strncat(color, temp, 6);
}

unsigned int rnd(unsigned int max)
{
	return rand() /(int)(((unsigned)RAND_MAX + 1) / max);
}

void randomcolor(int maxval, char color[7])
{
	int rgb;
	unsigned int r;

	color[0] = '#';
	/* ohne weiss strncat nicht, wo der string aufhoert */
	color[1] = '\0';

	for (rgb = 0; rgb < 3; rgb++)
	{
		r = rnd(maxval);

		append_color(r, color);
		printf("color: %d\n", r);
	}
	printf("fertig: %s\n", color);
}

/* alle 4 ecken angucken, durchschnitt der farben bilden */
/* oder: 2 ecken angucken, nur wenn nicht gleich durchschnitt */
/*
 * lieber doch nur 2 ecken. Sonst gibt's komische farben wenn in einer ecke 
 * z.B. ein dummes logo ist. Auch ausreichend.
 */
void lookupcolor(int w, int h, char color[7])
{
	Imlib_Color color_return;
	int rgb[3][3], border, i, temp;

	color[0] = '#';
	color[1] = '\0';

	/* links oben */
	imlib_image_query_pixel(0, 0, &color_return);
	rgb[0][0] = color_return.red;
	rgb[0][1] = color_return.green;
	rgb[0][2] = color_return.blue;

	/* links unten */
	imlib_image_query_pixel(0, h-1, &color_return);
	rgb[1][0] = color_return.red;
	rgb[1][1] = color_return.green;
	rgb[1][2] = color_return.blue;

	/* durchschnitt der ersten beiden ecken */
	for (i = 0; i < 3; i++) {
		temp = 0;
		for (border = 0; border < 2; border++)
			temp += rgb[border][i];

		temp = temp / 2;

		append_color(temp, color);
	}
}

/*
 * A program that wants to set the background should:
 *  - Create a persistant pixmap by calling to the server with:
 *        dpy = XOpenDisplay(display_name);
 *        XSetCloseDownDisplay (dpy, RetainPixmap);
 *        pmap = XCreatePixmap (dpy, DefaultRootWindow (dpy), width, height,
 *                              DefaultDepthOfScreen (DefaultScreenOfDisplay (dpy)));
 *        XCloseDisplay (dpy)
 *
 * - Grab the server
 * - See if ESETROOT_PMAP_ID is set
 * - If so, call XKillClient(xdisplay, *(Pixmap*)data_esetroot)
 * - Change ESETROOT_PMAP_ID and _XROOTPMAP_ID to point to the * new pixmap
 * - Set the pixmap as the background for the root window
 * - Ungrab the server
 *
 * http://mail.gnome.org/archives/wm-spec-list/2002-January/msg00006.html
 * http://www.eterm.org/docs/view.php?doc=ref
 *
 *
 * http://foo-projects.org/pipermail/xfce4-dev/2003-January/001703.html
 * > Or do you mean I should look for ESETROOT_PMAP_ID and destroy it when
 * > xfdesktop starts?
 * Yes, I think you should XKillClient the owner of  ESETROOT_PMAP_ID and
 * remove the property. 
 */
int setroot(Display *Xdisplay, Window Xroot, Screen *scr, int screen, char *color,
		int w, int h, int retain)
{
	Pixmap p = None, temp_pmap = None, m = None;
	GC gc;
	XGCValues gcv;
	XColor xcolor;
	int x, y;

	remove_property(Xdisplay, Xroot);

	p = XCreatePixmap(Xdisplay, Xroot, scr->width, scr->height, DefaultDepth(Xdisplay, screen));
	gcv.foreground = gcv.background = BlackPixel(Xdisplay, screen);

	if (color && XParseColor(Xdisplay, DefaultColormap(Xdisplay, screen), color, &xcolor)
			&& XAllocColor(Xdisplay, DefaultColormap(Xdisplay, screen), &xcolor)) {
		gcv.foreground = gcv.background = xcolor.pixel;
	}
	/* Create a Graphics Context TODO: fehler abfangen */
	gc = XCreateGC(Xdisplay, p, (GCForeground | GCBackground), &gcv);

	XFillRectangle(Xdisplay, p, gc, 0, 0, scr->width, scr->height);
	x = (scr->width - w) >> 1;
	y = (scr->height - h) >> 1;

	imlib_context_set_drawable(Xroot);
	imlib_render_pixmaps_for_whole_image_at_size(&temp_pmap, &m, w, h);

	if (temp_pmap != None) {
		XSetTile(Xdisplay, gc, temp_pmap);
		XSetTSOrigin(Xdisplay, gc, x, y);
		XSetFillStyle(Xdisplay, gc, FillTiled);
		XFillRectangle(Xdisplay, p, gc, x, y, w, h);

		/* comment from control-center (applier.c):
		 * Set the root pixmap, and properties pointing to it. We
		 * do this atomically with XGrabServer to make sure that
		 * we won't leak the pixmap if somebody else it setting
		 * it at the same time. (This assumes that they follow the
		 * same conventions we do)
		 *
		 * http://mail.gnome.org/archives/wm-spec-list/2002-January/msg00014.html:
		 * I've take a look at the Esetroot source code (0.9.1) and it
		 * does not grab/ungrab the server. This cause a program (e.g.,
		 * rxvt) which use parental relativity for pseudo-transparency
		 * and which monitor _XROOTPMAP_ID for background update to fail
		 * to update its background (as it receives the _XROOTPMAP_ID
		 * property before the background change).
		 */
		XGrabServer(Xdisplay);

		if (!retain)
			set_pixmap_property(Xdisplay, Xroot, p);
		else
			set_pixmap_permanent(Xdisplay, Xroot, p);

		XSetWindowBackgroundPixmap(Xdisplay, Xroot, p);
		XClearWindow(Xdisplay, Xroot);

		XUngrabServer(Xdisplay);

		/* send the request to the server */
		XFlush(Xdisplay);
		/* ?  XFreePixmap(Xdisplay, temp_pmap); */
		imlib_free_pixmap_and_mask(temp_pmap);
		CurrentPixmap = p;
	}

	XFreeGC(Xdisplay, gc);

	return 1;
}

int changeimage(Display *Xdisplay, Window Xroot, Screen *scr, int screen,
		const char *image, int retain)
{
	double maxscale = (double)cfg.maxscale / 100;
	int minborder = cfg.minborder;
	Imlib_Image im;
	int w, h;
	char color[8];

	double x_ratio, y_ratio;
	int *border, *full;

	/* load image */
	im = imlib_load_image_immediately(image);
	if (im == NULL) {
		return 0;
	}

	/* kill own last image */
	if (CurrentPixmap != None) {
		XFreePixmap(Xdisplay, CurrentPixmap);
	}

	imlib_context_set_image(im);

	w = imlib_image_get_width();
	h = imlib_image_get_height();

	/* border-color */
	lookupcolor(w, h, color);

	x_ratio = ((double)scr->width) / ((double)w);
	y_ratio = ((double)scr->height) / ((double)h);
	/* nicht zerren, kleinere ratio nehmen sonst laeuft's ueber */
	if (x_ratio > y_ratio) {
		/* hochkant */
		x_ratio = y_ratio;
		/* rand an dem frei bleibt */
		border = &h;
		full = &scr->height;
	} else {
		/* quer */
		/* rand an dem frei bleibt */
		border = &w;
		full = &scr->width;
	}

	/* don't make bigger than maxscale */
	if ((int) (*border * x_ratio) > (int) *border * maxscale)
		x_ratio = maxscale;

	/* ignore maxscale if the resulting border would be smaller minborder */
	if (*full - *border * x_ratio < minborder)
		x_ratio = ((double) *full) / ((double) *border);

	w = (int) (w * x_ratio);
	h = (int) (h * x_ratio);


	if (cfg.setbg) {
		setroot(Xdisplay, Xroot, scr, screen, color, w, h, retain);
	}
	
	if (cfg.command != NULL) {
		command(cfg.command, color, image, w, h);
	}
	/* Frees the current image in Imlib2's context AND removes it from the cache. */
	imlib_free_image_and_decache();
	return 1;

}
